;************************************************
;CTCSS Generator via PWM			* 
;H.E. 24.03.20					*
;CPU Clock 16MHz				*
;						*
;      Pinout :					*
;	RA0	PWM1				*
;	RA1	Channel Select (analog)		*
;	RA2	Status output	    		*
;	RA3	PTT (low active)		*
;************************************************
	list		N=70, ST=OFF, P=10F322
	#include	p10f322.inc
	__config	0xFCA0
;Port asignments
PWM	equ	0
ADC	equ	1
LED	equ	2
PTT	equ	3
;***************************************************************
;data register file
;***************************************************************
	cblock	0x40
outp_h		;PWM value high byte
outp_l		;PWM value low byte
accu_h		;24 bit accu
accu_m		;
accu_l		;
incr_h		;
incr_l		;Increment =2,097152s * f(tone)
cur_adc		;current ADC
old_adc		;previous ADC
temp		;division result
w_temp		;IRQ context saving
s_temp		;IRQ context saving
flag
	endc
;***************************************************************
;	code starts here
;***************************************************************
	org	0
	goto	p_init
;32s cycle
	org	4
	bcf	PIR1,TMR2IF
	movwf	w_temp
	swapf	STATUS,W
	movwf	s_temp		;save context
;update DAC
	movlw	high coslut
	movwf	PCLATH
	movf	accu_m,W
	btfss	accu_m,7	;pos or neg?
	goto	conv
	xorlw	0xFF		;negative slope
	addlw	1
conv	call	coslut		;get cosine value from table
	movwf	outp_h
	rrf	outp_h,F
	rrf	outp_l,W	;adjust byte order
	andlw	0x80
	movwf	PWM1DCL		;PWM low byte
	movf	outp_h,W
	andlw	0x7F
	movwf	PWM1DCH		;PWM high byte
;24 bit increment
	movf	incr_l,W
	addwf	accu_l,F
	btfss	STATUS,C
	goto	add1
	incf	accu_m,F
	btfsc	STATUS,Z
	incf	accu_h,F
add1	movf	incr_h,W
	addwf	accu_m,F
	btfsc	STATUS,C
	incf	accu_h,F
	btfss	accu_h,4
	goto	irq_ex
	bsf	flag,0		;read ADC every 250ms
	clrf	accu_h
;
irq_ex	swapf	s_temp,W
	movwf	STATUS
	swapf	w_temp,F
	swapf	w_temp,W	;restore context
	retfie
;***************************************************************
;start of programm 	initialize ports and peripherals
;***************************************************************
p_init	movlw	0x70
	movwf	OSCCON		;16MHz internal clock
;Ports
	movlw	0x0A
	movwf	TRISA		;0,2= out, 1,3=in
	movlw	2
	movwf	ANSELA		;1=analog in
	movlw	8
	movwf	WPUA		;3=PTT in
	movlw	0x88
	movwf	OPTION_REG	;Pull-up enable
;PWM output
	movlw	4
	movwf	T2CON		;enable timer 2 for PWM
	movlw	0x7F		;9 bit resolution
	movwf	PR2		;19.61kHz PWM freq.
	movlw	0xC0		;PWM on
	movwf	PWM1CON
	clrf	PWM1DCH
;ADC
	movlw	0xC7
	movwf	ADCON		;8us TAD, ch 1
;variables
	clrf	accu_l
	clrf	accu_m
	clrf	accu_h
	clrf	flag
	clrf	old_adc
	clrf	incr_h
	movlw	.200
	movwf	incr_l
;Interrupt
	bsf	PIE1,TMR2IE
	bsf	INTCON,PEIE
	bsf	INTCON,GIE
;***************************************************************
;main programm	
;***************************************************************
main 
	btfsc	PORTA,PTT
	bsf	TRISA,0		;PTT high, output off
	btfss	PORTA,PTT
	bcf	TRISA,0		;PTT active, output on
	btfss	flag,0		;start ADC
	goto	main
	btfss	flag,1		;ADC busy?
	bsf	ADCON,1		;if not, start ADC
	bsf	flag,1
	btfsc	ADCON,1		;ADC ready?
	goto	main
	clrf	flag
;
get_adc
	movf	ADRES,W		;read ADC
	movwf	cur_adc
	subwf	old_adc,W	;compare with previous value
	btfsc	STATUS,Z
	goto	main		;no change->exit
	movf	cur_adc,W
	movwf	old_adc		;set previous=current value
	call	div		;scale adc range to 38 tones
;
	movlw	high ad_lut	;set high byte table address
	movwf	PCLATH
	movf	temp,W
	andlw	0x7E
	movwf	temp
	call	ad_lut		;get table entry high byte
	movwf	incr_h
	movf	temp,W
	iorlw	1
	call	ad_lut		;get table entry low byte
	movwf	incr_l
	bcf	PORTA,LED
	btfsc	temp,1
	bsf	PORTA,LED	;LED on on even values
	goto	main
;
div	clrf	temp
d_loop	movlw	.3
	subwf	cur_adc,F
	btfss	STATUS,C
	return
	incf	temp,F
	goto	d_loop
;
	org	0x100
;cosine look-up table
coslut	addwf	PCL,F
	retlw	.255
	retlw	.254
	retlw	.254
	retlw	.254
	retlw	.254
	retlw	.254
	retlw	.253
	retlw	.253
	retlw	.252
	retlw	.251
	retlw	.251
	retlw	.250
	retlw	.249
	retlw	.248
	retlw	.247
	retlw	.246
	retlw	.245
	retlw	.244
	retlw	.242
	retlw	.241
	retlw	.240
	retlw	.238
	retlw	.236
	retlw	.235
	retlw	.233
	retlw	.231
	retlw	.230
	retlw	.228
	retlw	.226
	retlw	.224
	retlw	.222
	retlw	.219
	retlw	.217
	retlw	.215
	retlw	.213
	retlw	.210
	retlw	.208
	retlw	.206
	retlw	.203
	retlw	.201
	retlw	.198
	retlw	.195
	retlw	.193
	retlw	.190
	retlw	.187
	retlw	.185
	retlw	.182
	retlw	.179
	retlw	.176
	retlw	.173
	retlw	.170
	retlw	.167
	retlw	.164
	retlw	.161
	retlw	.158
	retlw	.155
	retlw	.152
	retlw	.149
	retlw	.146
	retlw	.143
	retlw	.140
	retlw	.137
	retlw	.134
	retlw	.131
	retlw	.128
	retlw	.124
	retlw	.121
	retlw	.118
	retlw	.115
	retlw	.112
	retlw	.109
	retlw	.106
	retlw	.103
	retlw	.100
	retlw	.97
	retlw	.94
	retlw	.91
	retlw	.88
	retlw	.85
	retlw	.82
	retlw	.79
	retlw	.76
	retlw	.73
	retlw	.70
	retlw	.68
	retlw	.65
	retlw	.62
	retlw	.60
	retlw	.57
	retlw	.54
	retlw	.52
	retlw	.49
	retlw	.47
	retlw	.45
	retlw	.42
	retlw	.40
	retlw	.38
	retlw	.36
	retlw	.33
	retlw	.31
	retlw	.29
	retlw	.27
	retlw	.25
	retlw	.24
	retlw	.22
	retlw	.20
	retlw	.19
	retlw	.17
	retlw	.15
	retlw	.14
	retlw	.13
	retlw	.11
	retlw	.10
	retlw	.9
	retlw	.8
	retlw	.7
	retlw	.6
	retlw	.5
	retlw	.4
	retlw	.4
	retlw	.3
	retlw	.2
	retlw	.2
	retlw	.1
	retlw	.1
	retlw	.1
	retlw	.1
	retlw	.1
	retlw	0
;
;frequency to increment value conversion
ad_lut	addwf	PCL,F
	retlw	0
	retlw	low .141
	retlw	0
	retlw	low .151
	retlw	0
	retlw	low .156
	retlw	0
	retlw	low .162
	retlw	0
	retlw	low .167
	retlw	0
	retlw	low .173
	retlw	0
	retlw	low .179
	retlw	0
	retlw	low .186
	retlw	0
	retlw	low .192
	retlw	0
	retlw	low .199
	retlw	0
	retlw	low .205
	retlw	0
	retlw	low .210
	retlw	0
	retlw	low .218
	retlw	0
	retlw	low .225
	retlw	0
	retlw	low .233
	retlw	0
	retlw	low .241
	retlw	0
	retlw	low .250
	retlw	1
	retlw	low .259
	retlw	1
	retlw	low .268
	retlw	1
	retlw	low .277
	retlw	1
	retlw	low .287
	retlw	1
	retlw	low .297
	retlw	1
	retlw	low .308
	retlw	1
	retlw	low .318
	retlw	1
	retlw	low .330
	retlw	1
	retlw	low .341
	retlw	1
	retlw	low .353
	retlw	1
	retlw	low .366
	retlw	1
	retlw	low .379
	retlw	1
	retlw	low .392
	retlw	1
	retlw	low .406
	retlw	1
	retlw	low .428
	retlw	1
	retlw	low .443
	retlw	1
	retlw	low .459
	retlw	1
	retlw	low .475
	retlw	1
	retlw	low .492
	retlw	1
	retlw	low .509
	retlw	2
	retlw	low .527
	retlw	2
	retlw	low .527
	retlw	2
	retlw	low .527
	retlw	2
	retlw	low .527
	retlw	2
	retlw	low .527
	retlw	2
	retlw	low .527
;
	end
